require "../vector"

module Gravity
  # A `Gravity::Body` simulates a heavy object that generate a significant warp of space time.
  # Several bodies can be combined in a single `Gravity::Field`.
  # A body exists in a space of N dimensions
  # and all vectors related to the body must have exactly N dimensions too.
  class Body(N)
    G = 6.6743*10**-11.0 # m³/(kg*s²)

    EARTH2 = Gravity::Body(2).new(mass: 5972200000000000000000000.0, position: Vector(Float64, 2).zero)
    EARTH3 = Gravity::Body(3).new(mass: 5972200000000000000000000.0, position: Vector(Float64, 3).zero)

    getter :mass, :position, :g, :name

    @@name_idx = 0
    def self.next_name
      @@name_idx += 1
      "default~#{@@name_idx}"
    end

    def initialize(@mass : Float64, @position : Vector(Float64, N) = Vector(Float64, N).zero, @g : Float64 = G, @name : String = Body(0).next_name)
    end

    def acceleration(position : Vector(Float64, N)) : Vector(Float64, N)
      return Vector(Float64, N).zero if position == @position
      orientation = @position - position
      distance = orientation.magnitude
      magnitude = @g * @mass / (distance ** 3)
      # original is magnitude = g * m / d² but since we also need to make
      # orientation a magnitude 1 vector to multiply it with magnitude later
      # I quicken the computation here with a single ** 3 operation rather than /d**2 then /d again
      orientation.mult! magnitude
      orientation
    end

    # Non-moving bodies don't accelerate
    def acceleration
      Vector(Float64, N).zero
    end

    # Non-moving bodies don't move
    def speed
      Vector(Float64, N).zero
    end
  end
end
