require "../engine/*"
require "../ui/*"
require "../frequency_loop"
require "../imgui_helper"

require "crsfml"
require "imgui"
require "imgui-sfml"

require "./physics_sandbox/*"

require "log"
Log.setup(:debug)

class Log
  def self.debug(str)
    Log.debug { str }
  end
end

class Game
  @window : SF::RenderWindow
  @delta_clock : SF::Clock
  @screen_center_position : Vector(Float64, 2)
  @center_position : Vector(Float64, 2)
  @joystick_moved : Vector(Float64, 2)
  @selected_projectable : Projectable(2)
  @click_coordinates : Vector(Float64, 2)?

  SECOND_IN_NANO = 1_000_000_000
  DEFAULT_ZOOM = 3/16

  def accuracy_frequency_nano_rate
    ((SECOND_IN_NANO * @timescale) // @framerate).to_u64
  end

  def draw(item : SF::VertexBuffer, transform : SF::Transform? = nil)
    states = SF::RenderStates.new
    if transform
      states.transform = transform
    end
    @window.draw(item, states)
  end

  def draw(item : SF::Shape, transform : SF::Transform? = nil)
    states = SF::RenderStates.new
    if transform
      states.transform = transform
    end
    @window.draw(item, states)
  end

  def default_viewport
    SF.float_rect(0.0, 0.0, 1.0, @width / @height)
  end

  def initialize(@framerate : Int32 = 60, @width : Int32 = 800, @height : Int32 = 600)
    @window = SF::RenderWindow.new(
      SF::VideoMode.new(width, height),
      "Sandbox",
    )
    @zoom = 1.0
    @joystick_moved = Vector[0.0, 0.0]
    @screen_center_position = Vector[width / 2, height / 2]
    @center_position = Vector[0.0, 0.0]

    @delta_clock = SF::Clock.new
    @time_passed = Time::Span.zero
    @timescale = 1.0
    @window_frequency = Time::Span.new(nanoseconds: SECOND_IN_NANO // @framerate)

    ImGui::SFML.init(@window)
    @window.framerate_limit = @framerate

    @view = SF::View.new
    @view.viewport = default_viewport # SF.float_rect(0.0, 0.0, 1.0, @width / @height)
    @view.center = {@center_position[0], @center_position[1]}
    @window.view = @view

    # connected = SF::Joystick.connected?(0)
    # # How many buttons does joystick #0 support?
    # buttons = SF::Joystick.get_button_count(0)
    # # Does joystick #0 define a X axis?
    # has_x = SF::Joystick.axis?(0, SF::Joystick::X)
    # # Is button #2 pressed on joystick #0?
    # pressed = SF::Joystick.button_pressed?(0, 2)
    # # What's the current position of the Y axis on joystick #0?
    # position = SF::Joystick.get_axis_position(0, SF::Joystick::Y)

    @click_coordinates = nil

    @projectables = Array(Projectable(2)).new
    Projectable(2).set_hook_new ->(projectable : Projectable(2)) {
      @projectables << projectable
    }

    sun_offset = Vector[3*10.0**11, 3*10.0**11]
    @bodies = {
      vessel: Gravity::MovingBody(2).new(name: "vessel", mass: 1.0, g: 0.1, position: Vector[10.0, 10.0]),
      star:   Gravity::MovingBody(2).new(name: "star", mass: 300000.0, g: 0.1, position: Vector[1000.0, 1000.0], speed: Vector[0.0, 0.0]),
      pla1:   Gravity::MovingBody(2).new(mass: 15000.0, g: 0.1, position: Vector[500.0, 500.0], speed: Vector[12.0, 0.0]),
      pla2:   Gravity::MovingBody(2).new(mass: 18000.0, g: 0.1, position: Vector[1000.0, 1500.0], speed: Vector[-6.0, -6.0]),
      pla3:   Gravity::MovingBody(2).new(mass: 12000.0, g: 0.1, position: Vector[1500.0, 1500.0], speed: Vector[-6.0, 0.0]),
    }
    @gravited = [
      @bodies[:pla1], @bodies[:pla2], @bodies[:pla3], @bodies[:star],
    ] of Gravity::Body(2)
    @graviting = [
      @bodies[:pla1], @bodies[:pla2], @bodies[:pla3], @bodies[:vessel],
    ] of Gravity::MovingBody(2)
    @vessel_p = Projectable(2).new(
      gravity_body: @bodies[:vessel],
      color: {255, 0, 250},
      size: 5,
    )
    @star = Projectable(2).new(
      gravity_body: @bodies[:star],
      color: {200, 200, 50},
      size: 20,
    )
    @pla1 = Projectable(2).new(
      gravity_body: @bodies[:pla1],
      color: {150, 50, 50},
      size: 10,
    )
    @pla2 = Projectable(2).new(
      gravity_body: @bodies[:pla2],
      color: {50, 50, 150},
      size: 10,
    )
    @pla3 = Projectable(2).new(
      gravity_body: @bodies[:pla3],
      color: {50, 150, 50},
      size: 10,
    )
    # @projectables = {
    #   @star, @pla1, @pla2, @pla3, @star, @vessel_p,
    # }
    @selected_projectable = @vessel_p
    @g = Gravity::Field(2).new(@gravited)
    @vessel = Vessel(2).new(name: "Starbird", gravity_body: @bodies[:vessel], gravity_field: @g)
    @vessel.navigation.zoom_ratio = DEFAULT_ZOOM
    @timer = Physics::Tick::Timer.new
    @ui_events_handler = UI::EventHandler.new
  end

  MAX_DISTANCE_TO_SELECT_PROJECTABLE = 15

  def handle_events
    joystick_moved_modified = false

    while event = @window.poll_event
      ImGui::SFML.process_event(@window, event)

      case event
      when SF::Event::Closed
        @window.close
        exit
      when SF::Event::Resized
        @width = event.width.to_i32
        @height = event.height.to_i32
        @view.viewport = default_viewport
        @center_position = Vector[0.0, 0.0]
        @view.center = {@center_position[0], @center_position[1]}
        @window.view = @view
      when SF::Event::KeyPressed
        Log.debug "KeyPressed #{event}"
      when SF::Event::MouseMoved
      when SF::Event::MouseButtonReleased
        clicked_on_imgui_item = ImGui.is_any_item_hovered || ImGui.is_any_item_active
        if !clicked_on_imgui_item
          # don't match if click on the buttons of ImGui
          # click_coord = @window.map_pixel_to_coords({event.x, event.y}, @view)
          # m = (proj.position - Vector[click_coord.x, click_coord.y]).magnitude
          # mmhhhh it does not work as expected, bugged
          event_position = Vector[event.x, event.y]
          nearest_proj = @projectables.map do |projectable|
            pixel = @window.map_coords_to_pixel({projectable.position[0], projectable.position[1]}, @view)
            {
              projectable: projectable,
              distance: (Vector[pixel.x, pixel.y] - event_position).magnitude,
            }
          end.sort_by!{ |tuple| tuple[:distance] }.first
          if nearest_proj[:distance] <= MAX_DISTANCE_TO_SELECT_PROJECTABLE
            @selected_projectable.outline_color = {0, 0, 0}
            @selected_projectable.outline_size = 0
            @selected_projectable = nearest_proj[:projectable]
            @selected_projectable.outline_color = {255, 0, 0}
            @selected_projectable.outline_size = 5
            pp "Select a new body #{@selected_projectable.name}"
            @click_coordinates = nil
          else
            @selected_projectable.outline_color = {0, 0, 0}
            @selected_projectable.outline_size = 0
            @click_coordinates = Vector[event.x.to_f64, event.y.to_f64]
          end
        end
      when SF::Event::MouseButtonEvent
        pp "MouseButtonEvent #{event}"
      when SF::Event::JoystickButtonPressed
        @joystick_moved.zero!              # reset 0
        joystick_moved_modified = :vector # force exact set
      when SF::Event::JoystickMoved
        case event.axis
        when SF::Joystick::Axis::U, SF::Joystick::Axis::X
          @joystick_moved[0] = event.position
          joystick_moved_modified = :vector
        when SF::Joystick::Axis::V, SF::Joystick::Axis::Y
          @joystick_moved[1] = event.position
          joystick_moved_modified = :vector
        when SF::Joystick::Axis::Z
          @joystick_moved[0] = event.position
          joystick_moved_modified = :slow
        when SF::Joystick::Axis::R
          @joystick_moved[0] = event.position
          joystick_moved_modified = :accelerate
        end
      else
        pp event
      end

      @ui_events_handler.handle(event)
    end

    if joystick_moved_modified == :vector
      @vessel.pilot.set_exact_acceleration_axis(@joystick_moved)
    elsif joystick_moved_modified == :accelerate
      # can double the vector each second
      Log.debug "accelerate #{@joystick_moved}"
      @vessel.pilot.accelerate_current_vector!(1.0 + (@joystick_moved[0] + 100) / 2 * 0.01)
    elsif joystick_moved_modified == :slow
      # can slow
      Log.debug "slow down #{@joystick_moved}"
      @vessel.pilot.accelerate_current_vector!(-1.0 - (@joystick_moved[0] + 100) / 2 * 0.01)
    end
  end

  def execute_loop
    while @window.open? # restart the loop when if it ends
      FrequencyLoop.new(@window_frequency).loop do |i|
        @window.clear(SF::Color::Black)
        handle_events
        ImGui::SFML.update(@window, @delta_clock.restart)
        execute(i)
        ImGui::SFML.render(@window)
        @window.display
      end
    end
  end

  def accelerate_by_tick(movable_body : Gravity::MovingBody)
    gravity_field = @g.acceleration(movable_body.position)
    acceleration = movable_body.acceleration + gravity_field
    Physics.move_by_one_tick!(@timer.tick, movable_body.position, movable_body.speed, acceleration)
  end

  include PhysicsSandboxDraw
  include PhysicsSandboxGrid
  include PhysicsSandboxCadran

  def execute(tick : Int)
    @vessel.execute_orders!

    @timer.timeskip!(nanoseconds: accuracy_frequency_nano_rate)
    @time_passed += accuracy_frequency_nano_rate.nanoseconds
    @vessel.pilot.apply_current_mode
    gravity_field = @g.acceleration(@bodies[:vessel].position)
    @graviting.each { |graviting| accelerate_by_tick(graviting) }

    draw_bridge(gravity_field)
    draw_bodies

    object_position =
      if @vessel.navigation.zoom_select == 0
        @vessel.navigation.zoom_position = @vessel.gravity_body.position
      elsif @vessel.navigation.zoom_select == 1
        @vessel.navigation.zoom_position = @star.gravity_body.position
      elsif @vessel.navigation.zoom_select == 2
        @vessel.navigation.zoom_position
      else
        @vessel.navigation.zoom_position = @screen_center_position
      end
    @center_position = object_position
    @view.set_center(
      @center_position[0],
      @center_position[1],
    )
    if @zoom != @vessel.navigation.zoom_ratio
      @view.zoom @zoom / @vessel.navigation.zoom_ratio
      @zoom = @vessel.navigation.zoom_ratio
      if @zoom == 1.0
        @view.viewport = default_viewport
      end
    end
    @window.view = @view

    @projectables.each { |projectable| draw(projectable.sf_shape) }
    draw_all_trajectories
    # draw_grid
    @vessel.navigation.cadran.radius.each do |radius|
      draw_cadran(center: @vessel.gravity_body.position, radius: radius)
    end
  end

  def draw_all_trajectories
    t = Trajectory.new(body: @vessel_p.gravity_body, gravity_field: @g)
    t_with_gravity = t.compute(steps: 1000, accuracy: 500.millisecond * @timescale)
    draw_trajectory({255, 0, 0}, t_with_gravity)

    t = Trajectory.new(body: @vessel_p.gravity_body, gravity_field: Gravity::Field(2).zero)
    t_without_gravity = t.compute(steps: 20, accuracy: 5.second * @timescale)
    draw_trajectory({255, 255, 0}, t_without_gravity)

    t = Trajectory.new(body: @pla1.gravity_body, gravity_field: @g)
    t_without_gravity = t.compute(steps: 100, accuracy: 10.second * @timescale)
    draw_trajectory({255, 0, 0}, t_without_gravity)

    t = Trajectory.new(body: @pla2.gravity_body, gravity_field: @g)
    t_without_gravity = t.compute(steps: 100, accuracy: 10.second * @timescale)
    draw_trajectory({0, 0, 255}, t_without_gravity)

    t = Trajectory.new(body: @pla3.gravity_body, gravity_field: @g)
    t_without_gravity = t.compute(steps: 100, accuracy: 10.second * @timescale)
    draw_trajectory({0, 255, 0}, t_without_gravity)
  end

  def draw_trajectory(color : Tuple(Int32, Int32, Int32), trajectory)
    trajectory.each do |point|
      sf_point = SF::CircleShape.new(2)
      sf_point.fill_color = SF.color(*color)
      sf_point.position = SF.vector2(point[0], point[1])
      draw(sf_point)
    end
  end
end

Game.new(framerate: 120, width: 800, height: 600).execute_loop
